package org.dasein.cloud.jclouds.rackspace;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Locale;

import org.apache.log4j.Logger;
import org.dasein.cloud.CloudException;
import org.dasein.cloud.InternalException;
import org.dasein.cloud.services.cdn.ContentDistributionServices;
import org.dasein.cloud.services.cdn.Distribution;
import org.jclouds.rackspace.cloudfiles.CloudFilesAsyncClient;
import org.jclouds.rackspace.cloudfiles.CloudFilesClient;
import org.jclouds.rackspace.cloudfiles.domain.ContainerCDNMetadata;
import org.jclouds.rackspace.cloudfiles.options.ListCdnContainerOptions;
import org.jclouds.rest.RestContext;

public class RackspaceCDN implements ContentDistributionServices {

	static private final Logger logger = Logger.getLogger(RackspaceCDN.class);
    
    private RackspaceCloud provider = null;
    
    public RackspaceCDN(RackspaceCloud provider) {
        this.provider = provider;
    }
	
	public String create(String directory, String name, boolean active, String ... cnames) throws InternalException, CloudException {
        logger.debug("enter - create(String, String, boolean, String...)");
        try {
            RestContext<CloudFilesClient,CloudFilesAsyncClient> ctx = provider.getCloudFilesContext();

            try {
                try {
                    if( !getClient(ctx).containerExists(directory) ) {
                        if( !getClient(ctx).createContainer(directory) ) {
                            throw new CloudException("Could not create container for distribution."); 
                        }
                    }
                }
                catch( RuntimeException e ) {
                    logger.error("Could not create container '" + directory + "': " + e.getMessage());
                    e.printStackTrace();
                    throw new CloudException(e);
                }
                if( active ) {
                    try {
                        getClient(ctx).enableCDN(directory);
                    }
                    catch( RuntimeException e ) {
                        logger.error("Could not enable distribution '" + directory + "': " + e.getMessage());
                        e.printStackTrace();
                        throw new CloudException(e);
                    }
                }
            }
            finally {
                ctx.close();
            }
            return directory;
        }
        finally {
            logger.debug("exit - create(String, String, boolean, String...)");            
        }
	}

	public void delete(String distributionId) throws InternalException, CloudException {
	    logger.debug("enter - delete(String)");
	    try {
            RestContext<CloudFilesClient,CloudFilesAsyncClient> ctx = provider.getCloudFilesContext();
	        
	        try {
	            getClient(ctx).disableCDN(distributionId);
	        }
	        finally {
	            ctx.close();
	        }
	    }
	    finally {
	        logger.debug("exit - delete(String)");
	    }
	}
    
    private CloudFilesClient getClient(RestContext<CloudFilesClient,CloudFilesAsyncClient> ctx) {
        return ctx.getApi();
    }
	    
	public Distribution getDistribution(String distributionId) throws InternalException, CloudException {
	    for( Distribution distribution : list() ) {
	        if( distribution.getProviderDistributionId().equals(distributionId) ) {
	            return distribution;
	        }
	    }
	    return null;
	}

	public String getProviderTermForDistribution(Locale locale) {
		return "container";
	}
	
	public Collection<Distribution> list() throws InternalException, CloudException {
	    logger.debug("enter - list()");
	    try {
	        ArrayList<Distribution> distributions = new ArrayList<Distribution>();
            Collection<ContainerCDNMetadata> list;
	        
	        try {
                RestContext<CloudFilesClient,CloudFilesAsyncClient> ctx = provider.getCloudFilesContext();

	            try {
	                list = getClient(ctx).listCDNContainers((ListCdnContainerOptions)null);
	            }
	            finally {
	                ctx.close();
	            }
	        }
	        catch( RuntimeException e ) {
	            logger.error("Could not list distributions in Rackspace: " + e.getMessage());
	            e.printStackTrace();
	            throw new CloudException(e);
	        }
	        for( ContainerCDNMetadata item : list ) {
	            if( item != null ) {
	                Distribution distribution = toDistribution(item);
	                
	                if( distribution != null ) {
	                    distributions.add(distribution);
	                }
	            }
	        }
	        return distributions;
	    }
	    finally {
	        logger.debug("exit - list()");
	    }
	}

	public void update(String distributionId, String name, boolean active, String ... cnames) throws InternalException, CloudException {
        logger.debug("enter - update(String, String, boolean, String ...)");
        try {
            Distribution distribution = getDistribution(distributionId);
            
            if( active ) {
                if( distribution == null ) {
                    create(name, name, active, cnames);
                }
                else if( !distribution.isDeployed() ) {
                    try {
                        RestContext<CloudFilesClient,CloudFilesAsyncClient> ctx = provider.getCloudFilesContext();

                        try {
                            getClient(ctx).enableCDN(distribution.getLocation());
                        }
                        finally {
                            ctx.close();
                        }
                    }
                    catch( RuntimeException e ) {
                        logger.error("Could not enable " + distribution.getProviderDistributionId() + ": " + e.getMessage());
                        e.printStackTrace();
                        throw new CloudException(e);
                    }
                }
            }
            else if( distribution != null && distribution.isDeployed() || distribution.isActive() ) {
                try {
                    RestContext<CloudFilesClient,CloudFilesAsyncClient> ctx = provider.getCloudFilesContext();

                    try {
                        getClient(ctx).disableCDN(distribution.getLocation());
                    }
                    finally {
                        ctx.close();
                    }
                }
                catch( RuntimeException e ) {
                    logger.error("Failed to disable " + distribution.getProviderDistributionId() + ": " + e.getMessage());
                    e.printStackTrace();
                    throw new CloudException(e);
                }
            }
        }
        finally {
            logger.debug("exit - update(String, String, boolean, String ...)");            
        }
	}

	private Distribution toDistribution(ContainerCDNMetadata rackspaceDistribution) {
	    Distribution distribution = new Distribution();
	    
        distribution.setName(rackspaceDistribution.getName());
	    distribution.setDnsName(rackspaceDistribution.getCDNUri().getHost());
	    distribution.setActive(rackspaceDistribution.isCDNEnabled());
	    distribution.setDeployed(rackspaceDistribution.isCDNEnabled());
	    distribution.setLocation(rackspaceDistribution.getName());
	    distribution.setProviderDistributionId(rackspaceDistribution.getName());
	    return distribution;
	}
}
