package org.dasein.cloud.jclouds.rackspace;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.List;
import java.util.Locale;
import java.util.Set;
import java.util.concurrent.Callable;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.Future;

import org.apache.log4j.Logger;
import org.dasein.cloud.AsynchronousTask;
import org.dasein.cloud.CloudException;
import org.dasein.cloud.InternalException;
import org.dasein.cloud.OperationNotSupportedException;
import org.dasein.cloud.services.image.ServerImage;
import org.dasein.cloud.services.server.Architecture;
import org.dasein.cloud.services.server.Platform;
import org.dasein.cloud.services.server.Server;
import org.dasein.cloud.services.server.ServerPersistence;
import org.dasein.cloud.services.server.ServerServices;
import org.dasein.cloud.services.server.ServerSize;
import org.dasein.cloud.services.server.ServerState;
import org.dasein.cloud.services.server.ServerStatistics;
import org.dasein.util.CalendarWrapper;
import org.dasein.util.Retry;
import org.jclouds.rackspace.cloudservers.CloudServersAsyncClient;
import org.jclouds.rackspace.cloudservers.CloudServersClient;
import org.jclouds.rackspace.cloudservers.domain.Addresses;
import org.jclouds.rackspace.cloudservers.domain.Flavor;
import org.jclouds.rackspace.cloudservers.domain.RebootType;
import org.jclouds.rackspace.cloudservers.domain.ServerStatus;
import org.jclouds.rackspace.cloudservers.options.CreateServerOptions;
import org.jclouds.rackspace.cloudservers.options.ListOptions;
import org.jclouds.rest.RestContext;

import com.google.common.base.Throwables;

public class RackspaceServer implements ServerServices {
	
	static private final Logger logger = Logger.getLogger(RackspaceServer.class);
    
    private RackspaceCloud provider = null;
    
    public RackspaceServer(RackspaceCloud provider) {
        this.provider = provider;
    }

    public void boot(String serverId) throws InternalException, CloudException {
        throw new OperationNotSupportedException("Rackspace does not support the pausing of servers.");
    }

    @Override
    public Server clone(String serverId, String intoDcId, String name, String description, boolean powerOn, String ... firewallIds) throws InternalException, CloudException {
        if( !powerOn ) {
            throw new OperationNotSupportedException("Rackspace does not support the pausing of servers.");
        }
        AsynchronousTask<ServerImage> task = provider.getImageServices().imageServer(serverId, null, null, name);
        Server server = getServer(serverId);
        ServerImage image;

        while( !task.isComplete() ) {
            try { Thread.sleep(10000L); }
            catch( InterruptedException e ) { }
        }
        if( task.getTaskError() != null ) {
            throw new CloudException(task.getTaskError());
        }
        image = task.getResult();
        while( !image.isAvailable() ) {
            try { Thread.sleep(15000L); }
            catch( InterruptedException e ) { }
            image = provider.getImageServices().getImage(image.getProviderServerImageId());
        }
        return launch(image.getProviderServerImageId(), new ServerSize(server.getSize()), intoDcId, name, null, null, false);
    }
    
    public Server define(String imageId, ServerSize size, String dataCenterId, String name, boolean withAnalytics, String ... firewalls) throws InternalException, CloudException {
        logger.debug("enter - define(String, ServerSize, String, String, boolean, String ...)");
        try {
            throw new OperationNotSupportedException("Rackspace launches servers on creation.");
        }
        finally {
            logger.debug("exit - define(String, ServerSize, String, String, boolean, String ...)");
        }
    }
    
    private CloudServersClient getClient(RestContext<CloudServersClient,CloudServersAsyncClient> ctx) {
        return ctx.getApi();
    }
    
    private CloudServersAsyncClient getAsyncClient(RestContext<CloudServersClient,CloudServersAsyncClient> ctx) {
        return ctx.getAsyncApi();
    }
    
	public String getConsoleOutput(String serverId, long sinceWhen) throws InternalException, CloudException {
	    return "";
	}

	public Collection<String> getFirewalls(String serverId) throws InternalException, CloudException {
        return new ArrayList<String>();
	}

	public String getProviderTermForServer(Locale locale) {
		return "server";
	}

	public Server getServer(String serverId) throws InternalException, CloudException {
	    for( Server server : list() ) {
	        if( server.getProviderServerId().equals(serverId) ) {
	            return server;
	        }
	    }
	    return null;
	}

    public ServerPersistence getServerPersistence() throws InternalException, CloudException {
        return ServerPersistence.ALL;
    }
    
    private ServerState getServerState(ServerStatus serverStatus) {
        switch( serverStatus ) {
            case ACTIVE: return ServerState.RUNNING;
            case BUILD: return ServerState.PENDING;
            case DELETED: return ServerState.TERMINATED;
            case ERROR: return ServerState.PAUSED;
            case REBOOT: return ServerState.REBOOTING;
            case SUSPENDED: return ServerState.PAUSED;
            default: return ServerState.REBOOTING;
        }
    }
    
	public ServerStatistics getServerStatistics(String serverId, long begin, long end) throws InternalException, CloudException {
		return new ServerStatistics();
	}
	
	static private long             lastSizeCacheTime = -1L;
	static private List<ServerSize> sizeCache         = null;
	static private ServerSize[]     sixtyFours        = null;
	
	ServerSize[] getSixtyFours() throws InternalException, CloudException {
        if( sixtyFours == null || (System.currentTimeMillis() >= (lastSizeCacheTime + (CalendarWrapper.MINUTE * 10L))) ) {
            getSupportedSizes(Architecture.I64);
        }
        if( sixtyFours == null ) {
            ServerSize[] tmp = new ServerSize[sizeCache.size()];
            int i = 0;
            
            for( ServerSize size : sizeCache ) {
                tmp[i++] = size;
            }
            return tmp;
        }
        return sixtyFours;
	}
	
	public Collection<ServerSize> getSupportedSizes(Architecture architecture) throws InternalException, CloudException {
        if( sizeCache == null || (System.currentTimeMillis() >= (lastSizeCacheTime + (CalendarWrapper.MINUTE * 10L))) ) {
    	    if( architecture.equals(Architecture.I64) ) {
                RestContext<CloudServersClient,CloudServersAsyncClient> ctx = provider.getCloudServersContext();
                ArrayList<ServerSize> sizes = new ArrayList<ServerSize>();
                ServerSize[] tmp = null;
                
    	        try {
        	        Collection<Flavor> flavors;
        	        
        	        try {
        	            flavors = getClient(ctx).listFlavors(ListOptions.Builder.withDetails());
        	        }
        	        catch( RuntimeException e ) {
        	            logger.error("Error retrieving flavors of " + architecture + " from Rackspace: " + e.getMessage());
        	            e.printStackTrace();
        	            throw new CloudException(e);
        	        }
        	        tmp = new ServerSize[flavors.size()];
        	        int i = 0;
        	        
                    for( Flavor flave : flavors ) {
                        ServerSize size = new ServerSize(String.valueOf(flave.getId()), flave.getName());
                        
                        sizes.add(size);
                        tmp[i++] = size;
                    }
    	        }
    	        finally {
    	            ctx.close();
    	        }
    	        lastSizeCacheTime = System.currentTimeMillis();
    	        sizeCache = Collections.unmodifiableList(sizes);
    	        if( tmp != null ) {
    	            sixtyFours = tmp;
    	        }
    	    }
    	    else {
    	        return Collections.emptyList();
    	    }
	    }
	    return sizeCache;
	}
	
    @Override
    public boolean isSupportsCloning() {
        return true;
    }
    
    public Server launch(String imageId, ServerSize size, String dataCenterId, String serverName, String withKey, String inVlanId, boolean withMonitoring, String... firewalls) throws InternalException, CloudException {
        logger.debug("enter - launch(String, ServerSize, String, String, String, String, boolean, String ...)");
        try {
            Future<org.jclouds.rackspace.cloudservers.domain.Server> task;
            
            try {
                RestContext<CloudServersClient,CloudServersAsyncClient> ctx = provider.getCloudServersContext();
                
                try {
                    task = getAsyncClient(ctx).createServer(serverName, Integer.parseInt(imageId), Integer.parseInt(size.getSizeId()), new CreateServerOptions[0]);
                }
                catch( Exception e ) {
                    throw new CloudException("Rackspace server launch failed: " + e.getMessage());
                }
                finally {
                    ctx.close();
                }
            }
            catch( RuntimeException e ) {
                logger.error("Could not create server " + imageId + "/" + size.getSizeId() + "/" + serverName + " in Rackspace: " + e.getMessage());
                e.printStackTrace();
                throw new CloudException(e);
            }
            if( task == null ) {
                return null;
            }
            org.jclouds.rackspace.cloudservers.domain.Server s;
            try {
                s = task.get();
                if( s == null ) {
                    return null;
                }
                return toServer(s);
            }
            catch( InterruptedException e ) {
                throw new CloudException("Server launch timed out: " + e.getMessage());
            }
            catch( ExecutionException e ) {
                throw new CloudException("Error launching server: " + e.getMessage());
            }
            catch( RuntimeException e ) {
                throw new CloudException("Unknown launch exception: " + e.getMessage());
            }
            catch( Error e ) {
                throw new CloudException("Unknown launch error: " + e.getMessage());
            }
        }
        finally {
            logger.debug("exit - launch(String, ServerSize, String, String, String, String, boolean, String ...)");
        }
	}

	public Collection<Server> list() throws InternalException, CloudException {
		logger.debug("enter - list()");
		try {
		    ArrayList<Server> servers = new ArrayList<Server>();
		    Collection<org.jclouds.rackspace.cloudservers.domain.Server> rawServers;
		    

            final RestContext<CloudServersClient,CloudServersAsyncClient> ctx = provider.getCloudServersContext();

            try {
                Retry<Collection<org.jclouds.rackspace.cloudservers.domain.Server>> retry = new Retry<Collection<org.jclouds.rackspace.cloudservers.domain.Server>>();

                rawServers = retry.retry(5, new Callable<Collection<org.jclouds.rackspace.cloudservers.domain.Server>>() {
                    @Override
                    public Collection<org.jclouds.rackspace.cloudservers.domain.Server> call() throws Exception {
                        return getClient(ctx).listServers(ListOptions.Builder.withDetails());
                    }
                    
                });
            }
            catch( Exception e ) {
                Throwable t = Throwables.getRootCause(e);
                
                e.printStackTrace();
                if( t != null ) {
                    throw new CloudException(t);
                }
                throw new CloudException(e);                    
            }
            finally {
                ctx.close();
            }

		    for( org.jclouds.rackspace.cloudservers.domain.Server server : rawServers ) { 
		        Server s = toServer(server);
		        
		        if( s != null ) {
		            servers.add(s);
		        }
		    }
		    return servers;
		}
		finally {
		    logger.debug("exit - list()");
		}
	}

	public void monitor(String serverId) throws InternalException, CloudException {
	    // NO-OP
	}


    public void pause(String serverId) throws InternalException, CloudException {
        throw new OperationNotSupportedException("Rackspace does not support the pausing of servers.");
    }

    public void reboot(String serverId) throws CloudException, InternalException {
        logger.debug("enter - reboot(String)");
        try {
            try {
                RestContext<CloudServersClient,CloudServersAsyncClient> ctx = provider.getCloudServersContext();

                try {
                    getClient(ctx).rebootServer(Integer.parseInt(serverId), RebootType.SOFT);
                }
                finally {
                    ctx.close();
                }
            }
            catch( RuntimeException e ) {
                logger.error("Rackspace failed to reboot the serve " + serverId + ": " + e.getMessage());
                e.printStackTrace();
                throw new CloudException(e);
            }
        }
        finally {
            logger.debug("exit - reboot(String)");
        }
    }
    
	public void stop(String serverId) throws InternalException, CloudException {
	    logger.debug("enter - stop(String)");
	    try {
	        try {
                RestContext<CloudServersClient,CloudServersAsyncClient> ctx = provider.getCloudServersContext();

                try {
                    getClient(ctx).deleteServer(Integer.parseInt(serverId));
                }
                finally {
                    ctx.close();
                }
	        }
	        catch( RuntimeException e ) {
	            logger.error("Rackspaced failed to terminate the server " + serverId + ": " + e.getMessage());
	            e.printStackTrace();
	            throw new CloudException(e);
	        }
	    }
	    finally {
	        logger.debug("exit - stop(String)");
	    }
	}
	
	private Server toServer(org.jclouds.rackspace.cloudservers.domain.Server rackspaceServer) {
	    Server server = new Server();
	    String str;
	    
	    server.setPersistent(true);
	    server.setArchitecture(Architecture.I64);
	    server.setCreateTime(0L);
        server.setDataCenterId("1");
        server.setRegionId("main");
	    server.setStartTime(0L);
	    server.setStopTime(-1L);
	    server.setTerminationTime(-1L);
	    str = rackspaceServer.getAdminPass();
	    if( str != null ) {
	        server.setRootPassword(str);
	    }
        server.setName(rackspaceServer.getName());
        server.setSize(String.valueOf(rackspaceServer.getFlavorId()));
	    server.setProviderServerId(String.valueOf(rackspaceServer.getId()));
	    server.setImageId(String.valueOf(rackspaceServer.getImageId()));
	    server.setCurrentState(getServerState(rackspaceServer.getStatus()));
	    server.setDescription(rackspaceServer.getMetadata().get("Server Label"));
	    if( server.getDescription() == null ) {
	        server.setDescription(rackspaceServer.getHostId());
	        if( server.getDescription() == null ) {
	            server.setDescription(server.getName());
	        }
	    }
	    Addresses addresses = rackspaceServer.getAddresses();
	    if( addresses != null ) {
            Set<String> list = addresses.getPrivateAddresses();
            String[] addrs = new String[list.size()];
            int i =0;
            
            for( String addr : list ) {
                addrs[i++] = addr;
            }
            server.setPrivateIpAddresses(addrs);
            
            list = addresses.getPublicAddresses();
            addrs = new String[list.size()];
            i = 0;
            for( String addr : list ) {
                addrs[i++] = addr;
            }
            server.setPublicIpAddresses(addrs);
	    }
	    server.setPlatform(Platform.guess(server.getName()));
	    if( server.getPlatform().equals(Platform.UNKNOWN) ) {
	        server.setPlatform(Platform.guess(server.getDescription()));
	        if( server.getPlatform().equals(Platform.UNKNOWN) ) {
	            server.setPlatform(Platform.UNIX);
	        }
	    }
	    return server;
	}
	
	public void unmonitor(String serverId) throws InternalException, CloudException {
	    // NO-OP
	}

    @Override
    public Collection<ServerStatistics> getServerStatisticsForPeriod(String serverId, long start, long end) throws InternalException, CloudException {
        return new ArrayList<ServerStatistics>();
    }  
}
